#include "agc.h"
#include <stdlib.h>
#include <stdio.h>

#ifndef FALSE
#define FALSE (1 == 0)
#endif

#ifndef TRUE
#define TRUE (0 == 0)
#endif

#define NEW(X) (X*) malloc(sizeof(X))
#define DEFNEW(NAME, TYPE) TYPE *NAME = NEW(TYPE)
#define MAKE_ARRAY(TYPE, N) (TYPE*) malloc((N) * sizeof(TYPE))

/* Free a agc_list structure. */
static void agc_free_list(struct agc_list *list) {
	struct agc_list *next;
	if(list) {
		next = list->next;
		free(list->obj);
		free(list);
		agc_free_list(next);
	}
}

/** Initialize AGC with the default heap size. */
struct agc *agc_init() {
	return agc_init_with_size(0x100000 / sizeof(AGC_WORD*));
}

/** Initialize AGC with the specified heap size. */
struct agc *agc_init_with_size(int size) {
	DEFNEW(agc, struct agc);

	if(agc) {
		agc->data = (AGC_WORD**) malloc(size * sizeof(AGC_WORD*));
		if(agc->data) {
			agc->size = size;
			agc->next = 0;
			agc->types = NULL;
		} else {
			free(agc);
			agc = NULL;
		}
	}

	return agc;
}

/** Finilize AGC. */
void agc_fini(struct agc *agc) {
	if(agc) {
		if(agc->data) free(agc->data);
		agc_free_list(agc->types);
		agc_free_list(agc->finalizers);
		free(agc);
	}
}

/** Allocate a num words of garbage-collected memory.
 * <root> is passed to agc_collect if the garbage collector is run.
 */
AGC_WORD *agc_alloc(struct agc *agc, int num, AGC_WORD **root) {
	if(agc->size >= agc->next + num) {
		agc->next += num;
		return agc->data + agc->next - num;
	} else {
		if(agc_collect(agc, root)) {
			return agc_alloc(agc, num, root);
		} else {
			return NULL;
		}
	}
}

/** Copy an object.
 * This function is intended to be called when agc already has a pointer
 * to newspace, but obj is still a pointer to oldspace.
 */
AGC_WORD *agc_copy_object(struct agc *agc, AGC_WORD *obj) {
	struct agc_list *types = agc->types;
	struct agc_type *type;

	while(types) {
		type = (struct agc_type*) types->obj;
		if(type->predicate(obj)) {
			return type->copy(agc, obj);
		}
		types = types->next;
	}

	fputs("FATAL ERROR - Invalid type tag encountered during collection\n",
		stderr);
	exit(0xff);
	return obj;
}

/* Iterate over the finalizers; run the ones that are applicable,
 * and update the ones that are not applicable.
 */
static void agc_run_finalizers(struct agc *agc) {
	struct agc_finalizer *finalizer;
	struct agc_list *next = agc->finalizers, *previous = NULL;

	while(next) {
		finalizer = (struct agc_finalizer*) next->obj;
		/* If finalizer->predicate returns a true value,
		 * the finalizer is run and discarded,
		 * otherwise, the finalizer is updated.
		 */
		if(finalizer->predicate(finalizer->obj)) {
			/* Run finalizer */
			finalizer->finalize(finalizer->obj);
			/* Discard finalizer */
			if(previous) {
				previous->next = next->next;
			} else {
				agc->finalizers = next->next;
			}
			free(finalizer);
			/* Save the pointer to the next finalizer
			 * before freeing its memory
			 */
			next = next->next;
			free(next);
		} else {
			/* Update finalizer */
			finalizer->update(finalizer);
			/* Set pointers for next iteration */
			previous = next;
			next = next->next;
		}
	}
}

/** Run the garbage collector.
 * <root> is a pointer to an object holding the rootset. Tracing and copying
 * will proceed from there, and the value will be updated with the new
 * location of the rootset.
 */
int agc_collect(struct agc *agc, AGC_WORD **root) {
	/* Save a pointer to oldspace. */
	AGC_WORD **oldspace = agc->data;

	/* Allocate memory for newspace. */
	agc->data = MAKE_ARRAY(AGC_WORD*, agc->size);
	if(agc->data) {
		agc->next = 0;
		*root = agc_copy_object(agc, *root);
		agc_run_finalizers(agc);
		return TRUE;
	} else {
		/* Failed to allocate memory for newspace */
		agc->data = oldspace;
		return FALSE;
	}
}

/** Register a type. */
int agc_register_type(struct agc *agc, struct agc_type *type) {
	DEFNEW(list, struct agc_list);

	if(list) {
		list->obj = type;
		list->next = agc->types;
		agc->types = list;
		return TRUE;
	} else {
		return FALSE;
	}
}

/** Register a new type.
 * predicate is run to determine of an object is of this type
 * copy is run to copy an object of this type to newspace
 * copy should also copy child objects
 */
struct agc_type *agc_register_new_type(struct agc *agc,
	int (*predicate) (AGC_WORD*),
	AGC_WORD *(*copy) (struct agc*, AGC_WORD*)) {

	DEFNEW(type, struct agc_type);

	if(type) {
		type->predicate = predicate;
		type->copy = copy;
		if(agc_register_type(agc, type)) return type;
		else {
			free(type);
			return NULL;
		}
	} else return NULL;
}

/** Add a finalizer. */
int agc_add_finalizer(struct agc *agc, struct agc_finalizer *finalizer) {
	DEFNEW(list, struct agc_list);

	if(list) {
		list->obj = finalizer;
		list->next = agc->finalizers;
		agc->finalizers = list;
		return TRUE;
	} else return FALSE;
}

/** Add a new finalizer.
 * obj is a pointer to the object that needs watching
 * predicate is run to determine if the object needs finalization
 * finalize is run to perform the actual finalization
 * update is run to update obj to point to the new location of the object
 */
struct agc_finalizer *agc_add_new_finalizer(struct agc *agc, AGC_WORD *obj,
	int (*predicate) (AGC_WORD*), void (*finalize) (AGC_WORD*),
	void (*update) (struct agc_finalizer*)) {

	DEFNEW(finalizer, struct agc_finalizer);
	
	if(finalizer) {
		finalizer->obj = obj;
		finalizer->predicate = predicate;
		finalizer->finalize = finalize;
		finalizer->update = update;
		if(agc_add_finalizer(agc, finalizer)) return finalizer;
		else {
			free(finalizer);
			return NULL;
		}
	} else return NULL;
}
